/********************************************************************

HashCart 2 - simple idea of turning GTEA into hash function

Runs with 128 key and 128 bit block without KW.

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

typedef unsigned char u8;
typedef unsigned long u32;
typedef long int s32;

/************** Full key size 128 - 4096 ***************************/
#define keysizebits  128
/*******************************************************************/

#define keysizebytes keysizebits  / 8
#define keysize keysizebytes / 4

/************** Number of algorithm full cycles ********************/
#define cycles 32
/*******************************************************************/

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

#define sboxsize 2 << sizeshift

static const u32 sbselect = 31 - sizeshift;

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d) (ROL(sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255], 13))

#define F(x) ((x) + ((x << 6) ^ (x >> 8)))

u32 roundkeys[2][cycles << 2];



void
setup (u32 id, u32 * key)
{
  u32 delta = 0x2CAD9447;
  u32 cdelta = 0x213B4680;
  u32 cnt = delta;
  u32 i;
  u32 kcnt = cdelta;

  for (i = 0; i != cycles << 2; i++)
    {

      roundkeys[id][i] = key[i % keysize];
      kcnt += key[i % keysize] + cdelta;
      roundkeys[id][i] += cnt;
      roundkeys[id][i] =
        subst (roundkeys[id][i], i % sboxsize, cnt % sboxsize,
               roundkeys[id][i] % sboxsize, kcnt % sboxsize);
      kcnt += roundkeys[id][i];
      cnt += delta;

    }






}


void
crypt (u32 id, u32 * data, u32 * out)
{
  u32 i, k, n;
  register u32 a, b, c, d, t;



  a = data[0];
  b = data[1];
  c = data[2];
  d = data[3];


  for (i = 0, n = 0, k = cycles << 1; i != cycles; i++, n++, k++)
    {
      a += ROR (F (b) + roundkeys[id][n], c >> 27);
      n++;
      b += ROR (F (a) + roundkeys[id][n], d >> 27);
      c ^=
        subst (ROR (F (a) + roundkeys[id][k], b >> 27), k % sboxsize,
               i % sboxsize, a >> sbselect, d >> sbselect);
      k++;
      d ^=
        subst (ROR (F (b) + roundkeys[id][k], a >> 27), c >> sbselect,
               b >> sbselect, k % sboxsize, n % sboxsize);

      t = a;
      a = c;
      c = t;
      t = b;
      b = d;
      d = t;

    }



  out[0] = a;
  out[1] = b;
  out[2] = c;
  out[3] = d;


}




u32 blocks = 0, currentblock = 0, blocksleft = 0;

u8 key[16];

FILE *stream;

void
InitHash (u32 datalen)
{


  if (datalen < 16)
    {
      blocks = 1;
    }
  else
    {
      blocks = (datalen + (datalen % 16)) / 16;
    }

  currentblock = 0;

  blocksleft = blocks;

  key[0] = 0xac;
  key[1] = 0xee;
  key[2] = 0xd0;
  key[3] = 0xbb;
  key[4] = 0x11;
  key[5] = 0xc2;
  key[6] = 0xc2;
  key[7] = 0x7b;
  key[8] = 0x02;
  key[9] = 0x35;
  key[10] = 0xda;
  key[11] = 0xf7;
  key[12] = 0xeb;
  key[13] = 0x35;
  key[14] = 0xaa;
  key[15] = 0x90;


}


void
Update (u8 * buffer)
{

  u32 i;
  u8 out[16];

  currentblock++;
  blocksleft--;


  setup (0, (u32 *) & key[0]);
  crypt (0, (u32 *) & buffer[0], (u32 *) & out[0]);

  for (i = 0; i != 16; i++)
    {
      key[i] = out[i] ^ buffer[i];
    }


}


int
main (int argc, char *argv[])
{

  u8 buffer[16];
  u32 fsize;

  int i, n;

  if (argc < 2)
    {
      fprintf (stderr, "usage: %s <input filename>\n", argv[0]);
      return (1);
    }

  if ((stream = fopen (argv[1], "rb+")) == NULL)
    {
      fprintf (stderr, "failed to open '%s' for reading.\n", argv[1]);
      return (1);
    }


  fseek (stream, 0, SEEK_END);
  fsize = ftell (stream);
  fseek (stream, 0, SEEK_SET);

  InitHash (fsize);

  do

    {


      memset (buffer, 0, 16);
      n = fread (buffer, 1, 16, stream);


      if (n != 0)
        {


          if (n != 16)
            {
              buffer[15] = n;

              if (n < 15)
                {
                  buffer[14] = 16 - n;
                }

            }


          Update (buffer);
        }

    }

  while (n == 16);




  fclose (stream);

  for (i = 0; i != 16; i++)
    {

      fprintf (stdout, "%2d  0x%x \n", i, key[i]);

    }

  fprintf (stdout, "\n");

  return (0);
}
